# -*- coding: utf-8 -*-
"""
----------------------------------------------------------
-------------------------- PBESA -------------------------
----------------------------------------------------------

@autor AKEN
@version 4.0.0
@date 05/02/25
"""

# --------------------------------------------------------
# Define resources
# --------------------------------------------------------

import datetime
from .templates import AFIRMATIVE_RESPONSE, GET_DATE

#------------------------------------------
# Define enumeration with states of the 
# dialog
#------------------------------------------

class DialogState:
    """ Define the dialog states """
    START = "start"
    ASK_CONFIRMATION_NEGATIVE = "ask confirmation negative"
    ASK_CONFIRMATION_POSITIVE = "ask confirmation_positive"
    END = "end"

#------------------------------------------
# Define Node class
#------------------------------------------

class ActionNode:
    """ Node class """

    def __init__(self, owner, performative, action, is_terminal=False):
        self.owner = owner
        self.children = {}
        self.action = action
        self.is_terminal = is_terminal
        self.performative = performative

    def add_child(self, tag, node):
        self.children[tag] = node
    
class DeclarativeNode:
    """ Node class """

    def __init__(self, owner, performative, text, is_terminal=False):
        self.text = text
        self.owner = owner
        self.children = {}
        self.is_terminal = is_terminal
        self.performative = performative

    def add_child(self, tag, node):
        self.children[tag] = node
    
class TerminalNode:
    """ Node class """

    def __init__(self, owner, performative):
        self.owner = owner
        self.performative = performative

class Node:
    """ Node class """

    def __init__(self, owner, performative, text=None, action=None, is_terminal=False, is_expert=False, value={}):
        self.text = text
        self.owner = owner
        self.value = value
        self.children = {}
        self.action = action
        self.is_expert = is_expert
        self.is_terminal = is_terminal
        self.performative = performative

    def add_child(self, tag, node):
        self.children[tag] = node

    def set_value(self, key, value):
        self.value[key] = value

    def run(self, dto):
        if self.text and not self.action:
            return self
        elif self.action:
            window = self.action(self, dto)
            if window.is_terminal:
                return window.action(window, dto) if window.action else window
            elif self.is_expert:
                return self
            return window.run(dto)
        raise Exception("Se ha estructurado mal el dialogo.")
    
    def generate_response(self):
        response = get_response_message()
        response['text'] = self.text
        if self.value:
            for key, value in self.value.items():
                response[key] = value
        return response

#------------------------------------------
# Define functions
#------------------------------------------

def get_response_message():
    """ Define the response message """
    return {
        "text": None,
        "share": False
    }

def check_yes(self, dto):
    """ Check if text is yes """
    user_response = dto['query']
    user_response = user_response.lower()
    if user_response == "si" or user_response == "sí" or user_response == "yes" or user_response == "s ":
        return self.children['yes']
    model_handler = self.owner.model_handler
    prompt = AFIRMATIVE_RESPONSE % user_response
    response = model_handler.make_request(prompt)
    if response:
        if "SI_ES" in response:
            return self.children['yes']
    return self.children['no']

def check_date(self, dto):
    """ Check if date is valid """
    date = dto['query']
    session = dto['session']
    model_handler = self.owner.model_handler
    # Get current date in %d-%m-%Y format.
    current_date = datetime.datetime.now().strftime("%d-%m-%Y")
    date_prompt = GET_DATE
    prompt = date_prompt.format(
        fecha_actual= current_date,
        texto= date
    )
    response = model_handler.make_request(prompt, engine="gpt-4")
    # Check if response is valid date.
    if response:
        try:
            reminder_date = datetime.datetime.strptime(response, '%Y-%m-%d %H:%M:%S')
            dto['reminder_date'] = reminder_date
            return self.children['yes']
        except:
            pass
    return self.children['no']


# -*- coding: utf-8 -*-
"""
----------------------------------------------------------
-------------------------- PBESA -------------------------
----------------------------------------------------------

@autor AKEN
@version 4.0.0
@date 08/08/24
"""

# --------------------------------------------------------
# Define resources
# --------------------------------------------------------

from abc import ABC, abstractmethod
from pbesa.models import AIFoundry, AzureInference, GPTService, ServiceProvider
from pbesa.social.dialog import DialogState, ActionNode, DeclarativeNode, TerminalNode

# --------------------------------------------------------
# Define common functions
# --------------------------------------------------------

def define_service_provider(provider, ai_service=None) -> None:
        # Define provider
        service = None
        service_provider = ServiceProvider()
        if "GPT" in provider:
            service = GPTService()
            service_provider.register("GPT", service)
        elif "AI_FOUNDRY" in provider:
            service = AIFoundry()
            service_provider.register("AI_FOUNDRY", service)
        elif "CUSTOM_ML" in provider:
            service = ai_service()
            service_provider.register("CUSTOM_ML", service)
        elif "AZURE_INFERENCE" in provider:
            service = AzureInference()
            service_provider.register("AZURE_INFERENCE", service)
        return service_provider, service

# --------------------------------------------------------
# Define Model component
# --------------------------------------------------------

class Model(ABC):
    """ Model class """

    def __init__(self) -> None:
        """ Constructor method """
        self.model:any = None
    
    def get_model(self) -> any:
        """ Get model method 
        :return: model
        """
        return self.model
    
    @abstractmethod
    def load_model(self, config:dict) -> None:
        """ Load model method
        :param config: config
        """ 
        pass

    @abstractmethod
    def train_model(self, model_config:dict, data:any) -> any:
        """ Train model method
        :param model_config: model_config
        :param data: data
        :return: any
        """
        pass

    @abstractmethod
    def evaluate_model(self, data:any) -> any:
        """ Validate model method
        :param data: data
        :return: any
        """
        pass

    @abstractmethod
    def fit_model(self, model_config:dict, data:any) -> any:
        """ Fit model method
        :param model_config: model_config
        :param data: data
        :return: any
        """
        pass

    @abstractmethod
    def predict(self, data:any) -> any:
        """ Predict method
        :param data: data
        :return: any
        """
        pass

# --------------------------------------------------------
# Define Generative component
# --------------------------------------------------------

class Generative(ABC):
    """ Generative class """

    def __init__(self) -> None:
        """ Constructor method """
        self.model:any = None
    
    def get_model(self) -> any:
        """ Get model method 
        :return: model
        """
        return self.model
    
    def set_model(self, model:any) -> None:
        """ Set model method
        :param model: model
        """ 
        self.model = model

    def generate(self, data:any, config:dict) -> any:
        """ Generate method
        :param data: data
        :param config: config
        :return: any
        """
        pass

# --------------------------------------------------------
# Define Augmented Generation component
# --------------------------------------------------------

class AugmentedGeneration(ABC):
    """ Augmented Generation """

    def __init__(self) -> None:
        """ Constructor method """
        self.model:any = None
        self.model_conf:dict = None
        self.__work_memory:list = []
        # Define role
        role = self.define_role() 
        self.__role = role.strip() if role else "Undefined"
        # Define the provider
        self.__service_provider = None
        # Define AI service
        self.__ai_service = None
        # Set up model
        self.set_up_model()

    def set_up_model(self):
        """ Set up model method """
        # Define role
        self.__work_memory.append({"role": "system", "content": self.__role})
        
    def get_model(self) -> any:
        """ Get model method 
        :return: model
        """
        return self.model
    
    def load_model(self, provider, config, ai_service=None) -> None:
        self.__service_provider, service = define_service_provider(provider, ai_service)
        service.setup(config, self.__work_memory)
        self.__ai_service = service
    
    def update_world_model(self, fact:str) -> None:
        """ Update world model method
        :param fact: fact
        :return: str
        """
        self.__work_memory.append({"role": "user", "content": fact})

    def reset(self) -> None:
        """ Reset method """
        self.__work_memory = []
        # Set up model
        self.set_up_model()

    def derive(self, query) -> str:
        """ Generate method
        :return: str
        """
        content = self.retrieval(query)
        self.__work_memory.append({"role": "system", "content": f"A partir de la siguiente información: {content} responde a la siguiente consulta:"})
        self.__work_memory.append({"role": "user", "content": query})
        return self.__ai_service.generate()

    @abstractmethod
    def define_role(self) -> str:
        """ Set role set method """
        pass

    @abstractmethod
    def retrieval(self, query) -> str:
        """ Set retrieval method
        :param query: query
        :return: str
        """
        pass

# --------------------------------------------------------
# Define Rational component
# --------------------------------------------------------

class Rational(ABC):
    """ Rational class """

    def __init__(self) -> None:
        """ Constructor method """
        self.model:any = None
        self.__work_memory:list = []
    
    def get_model(self) -> any:
        """ Get model method 
        :return: model
        """
        return self.model
    
    def set_model(self, model:any) -> None:
        """ Set model method
        :param model: model
        """ 
        self.model = model
    
    def derive(self, data:any, config:dict) -> any:
        """ Generate method
        :param data: data
        :param config: config
        :return: any
        """
        pass

# --------------------------------------------------------
# Define Production component
# --------------------------------------------------------

class Production(ABC):
    """ Rational class """

    def __init__(self) -> None:
        """ Constructor method """
        self.model:any = None
        self.model_conf:dict = None
        self.__work_memory:list = []
        # Define role
        role = self.define_role() 
        self.__role = role.strip() if role else "Undefined"
        # Define rule set
        self.__rule_set = self.define_rule_set()
        # Define example
        self.__example = self.define_example()
        # Define model adapter
        self.__model_adapter = None
        # Set up model
        self.set_up_model()

    def set_up_model(self):
        """ Set up model method """
        # Define role
        self.__work_memory.append({"role": "system", "content": self.__role})
        # Iterate over the rule set
        self.__work_memory.append({"role": "system", "content": "Sigue estrictamente las siguientes reglas:"})
        for rule in self.__rule_set:
            self.__work_memory.append({"role": "system", "content": rule})
        self.__work_memory.append({"role": "system", "content": "Ejemplo:"})
        # Define example
        self.__work_memory.append({"role": "system", "content": self.__example})
        
    def get_model(self) -> any:
        """ Get model method 
        :return: model
        """
        return self.model
    
    def load_model(self, model:any, model_conf:dict) -> None:
        """ Set model method
        :param model: model
        :param model_conf: Configuration
        """
        self.model = model
        self.model_conf = model_conf
        #self.__model_adapter = OpenAIAdapter(self.model, self.model_conf, self.__work_memory)

    def get_rule_set(self) -> list:
        """ Get rule set method
        :return: list
        """
        return self.__rule_set
    
    def update_world_model(self, fact:str) -> None:
        """ Update world model method
        :param fact: fact
        :return: str
        """
        self.__work_memory.append({"role": "user", "content": fact})

    def reset(self) -> None:
        """ Reset method """
        self.__work_memory = []
        # Set up model
        self.set_up_model()

    def derive(self) -> str:
        """ Generate method
        :return: str
        """
        return self.__model_adapter.generate()

    @abstractmethod
    def define_role(self) -> str:
        """ Set role set method """
        pass

    @abstractmethod
    def define_rule_set(self) -> tuple:
        """ Set rule set method """
        pass

    @abstractmethod
    def define_example(self) -> str:
        """ Set example set method """
        pass

# --------------------------------------------------------
# Define Dialog component
# --------------------------------------------------------

class Dialog(ABC):
    """ Dialog """

    def __init__(self) -> None:
        """ Constructor method """
        self.model:any = None
        self.model_conf:dict = None
        self.__work_memory:list = []
        # Define role
        role = self.define_role() 
        self.__role = role.strip() if role else "Undefined"
        # Define DFA
        dfa = self.define_dfa()
        self.__dfa = dfa if dfa else "Undefined"
        # Set dialog state
        self.__dialog_state = DialogState.START
        # Define the provider
        self.__service_provider = None
        # Define AI service
        self.__ai_service = None
        # Set up model
        self.set_up_model()
        

    def set_up_model(self):
        """ Set up model method """
        # Define role
        self.__work_memory.append({"role": "system", "content": self.__role})
        
    def get_model(self) -> any:
        """ Get model method 
        :return: model
        """
        return self.model
    
    def set_dialog_state(self, dialog_state) -> None:
        """ Set dialog state method
        :param dialog_state: dialog_state
        """
        self.__dialog_state = dialog_state
    
    def get_dialog_state(self) -> DialogState:
        """ Get dialog state method
        :return: DialogState
        """
        return self.__dialog_state
    
    def load_model(self, provider, config, ai_service=None) -> None:
        self.__service_provider, service = define_service_provider(provider, ai_service)
        service.setup(config, self.__work_memory)
        self.__ai_service = service
    
    def update_world_model(self, fact:str) -> None:
        """ Update world model method
        :param fact: fact
        :return: str
        """
        self.__work_memory.append({"role": "user", "content": fact})

    def reset(self) -> None:
        """ Reset method """
        self.__work_memory = []
        # Set up model
        self.set_up_model()

    def transition(self, dialog_state, query) -> str:
        """ Generate method
        :return: str
        """
        text = ""
        path_tag = None
        node = self.__dfa[dialog_state]
        if isinstance(node, DeclarativeNode):
            text = node.text
            self.__work_memory.append({"role": "system", "content": text})
            self.__work_memory.append({"role": "user", "content": query})
        elif isinstance(node, ActionNode):
            node, text = node.action(node, query)
            print("-> node:", node.performative)
            print("-> text:", text)
            self.__work_memory.append({"role": "user", "content": query})
            self.__work_memory.append({"role": "system", "content": text})
            self.__work_memory.append({"role": "system", "content": text})
            new_owner = node.owner
            new_dialog_state = node.performative
            return new_owner, new_dialog_state, self.__ai_service.generate()
        path_tag = next((tag for tag in ['do', 'yes', 'no'] if tag in node.children), None)
        print("path_tag", path_tag)       
        if path_tag:
            new_owner = node.children[path_tag].owner
            new_dialog_state = node.children[path_tag].performative
            if not isinstance(node, TerminalNode):
                print("=> new_owner:", new_owner, "new_dialog_state:", new_dialog_state)
                return new_owner, new_dialog_state, self.__ai_service.generate()
            print("=> new_owner:", new_owner, "new_dialog_state:", new_dialog_state)
            return new_owner, new_dialog_state, None
        else:
            raise Exception("No path found")

    @abstractmethod
    def define_role(self) -> str:
        """ Set role set method """
        pass

    @abstractmethod
    def define_dfa(self) -> dict:
        """ Set role set method """
        pass